﻿'3D-Type-Defs and Helper-Classes for 3DSurfaceDemo.bas... the only dependencies to the RC5_Objects.bi-module and
'the RC5-libs are in the Drawing-Method of c3DQuad and the little cSpectrum-Class... the rest is generically usable

Public Type t3D
    x As Double
    y As Double
    z As Double
End Type

Public Type t2D
    x As Double
    y As Double
End Type

Enum eFuncType 'this enum is usded in the Struct below, to apply matching formulas for the z-Coord
    UnitCube
    GaussianBell
    Rosenbrock
End Enum

Public Type tAreaInfo
    FType  As eFuncType
    Area3D(Any, Any) As t3D '<- Area3D is dx/dy-sized externally and then filled, based on the eFuncType above

    'these Members describe Info, filled within Procedure: Normalize3DDataAndFillQuads (further below in this module)
    As Double xMin, xMax, xScale 
    As Double yMin, yMax, yScale 
    As Double zMin, zMax, zScale
End Type
 
Public Enum eBackFaceStyle 'used for Quad-Renderings
    Backface_normal
    Backface_greyed
    BackFace_none
End Enum

'a few 3D-Vector-Functions (as needed for e.g. the Normal-Calculations in c3DQuad)
Function NewV(x As Double, y As Double, z As Double) As t3D
    Dim V As t3D = (x, y, z): Return V
End Function
Function SubV(A As t3D, B As t3D) As t3D
    Dim V As t3D = (A.x - B.x, A.y - B.y, A.z - B.z): Return V
End Function
Function CrossV(A As t3D, B As t3D) As t3D
    Dim V As t3D = (A.y * B.z - A.z * B.y, A.z * B.x - A.x * B.z, A.x * B.y - A.y * B.x): Return V
End Function
Function MagV(V As t3D) As Double
    Return Sqr(V.x * V.x + V.y * V.y + V.z * V.z)
End Function

'*** ClassDef which ensures a Color-Spectrum for fast Lookups over: Spectrum.Lut(Index) ... efficiently done per cairo-colorstop-patterns
Type cSpectrum 
    LUT(0 To 1000) As Long '<- define, over how many ColorValues we want to spread the spectrum
    LUTSrf As RC5_CairoSurface 'holds the accompanying Cairo-Surface (for convenient rendering of the Spectrum-Bar in the cpPlot-Panel)
    Declare Constructor()
End Type
Constructor cSpectrum() 'constructs a spectrum in a Color-LookupTable
    Dim Stops(0 To 4) As Long = {vbBlue, vbMagenta, vbRed, vbYellow, vbGreen} 'define the Base-Stop-Colors as you like
    Dim Pat As RC5_CairoPattern = RC5_CairoPattern(0, 0, 0, UBound(LUT)) 
    For i As Long = 0 To UBound(Stops): Pat.AddColorStop i / UBound(Stops), Stops(i): Next

    LUTSrf = RC5_CairoSurface(1, UBound(LUT)+1)
    LutSrf.CreateContext.Paint 1, Pat
    Dim B As UByte Ptr = LutSrf.DataPtr
    For i As Long = 0 To UBound(LUT) 'fill the LUT-Array (inverting from BGR-Order to RGB-Order)
       LUT(i) = (B[i*4+0] SHL 16) Or (B[i*4+1] SHL 8) Or B[i*4+2]
    Next
End Constructor
Dim Shared Spectrum As cSpectrum 'we hold it available here already as a singleton, for anybody who needs it
'*** End of cSpectrum


'*** ClassDef for Projections of a 3D-Point onto a 2D-Plane
Type cProjection 
    As Double PerspectiveDist, FocalLength, AzimuthDeg, AltitudeDeg
    Declare Constructor()
    Declare Sub UpdateAngles(AzimDeg As Double, AltiDeg As Double)
    Declare Sub Calc2DPt(x As Double, y As Double, z As Double, ByRef px As Double, ByRef py As Double, ByRef Depth As Double=0)
    Private:
    As Double SinTh, CosTh, SinPh, CosPh, CosThSinPh, SinThSinPh, CosThCosPh, SinThCosPh
End Type
Constructor cProjection() 
    PerspectiveDist = 8: FocalLength = 800
End Constructor
'since we have a Class here, we can pre-calculate efficiently (on external Angle-Changes) ...
Sub cProjection.UpdateAngles(AzimDeg As Double, AltiDeg As Double) 
    If AzimDeg > 360 Then AzimDeg = 360 Else If AzimDeg < 0 Then AzimDeg = 0
    If AltiDeg > 180 Then AltiDeg = 180 Else If AltiDeg < -180 Then AltiDeg = -180
    AzimuthDeg = AzimDeg: AltitudeDeg = AltiDeg

    SinTh = Sin((AzimuthDeg - 180) * 1.74532925199433E-02)
    CosTh = Cos((AzimuthDeg - 180) * 1.74532925199433E-02)

    CosPh = Cos(AltitudeDeg * 1.74532925199433E-02)
    SinPh = Sin(AltitudeDeg * 1.74532925199433E-02)

    CosThSinPh = CosTh * SinPh: SinThSinPh = SinTh * SinPh
    CosThCosPh = CosTh * CosPh: SinThCosPh = SinTh * CosPh
End Sub
'...so that this "workhorse-routine" doesn't have to call any expensive Sin/Cos anymore, when all the Polygon-Points are re-calculated
Sub cProjection.Calc2DPt(x As Double, y As Double, z As Double, ByRef px As Double, ByRef py As Double, ByRef Depth As Double=0) '
    Dim x_ As Double, y_ As Double, z_ As Double, Divisor As Double
    'get the viewpoint (px, py) in eye-coordinates from the 3D-Pt (x,y,z)
    x_ = -y * SinTh + x * CosTh
    y_ = -y * CosThSinPh - x * SinThSinPh + z * CosPh
    z_ = -y * CosThCosPh - x * SinThCosPh - z * SinPh
    Depth = z_ + 0.5 / PerspectiveDist

    '2D-Pt in plane
    Divisor = z_ + PerspectiveDist: If Divisor = 0 Then Divisor = 1E-20
    px = -FocalLength * x_ / Divisor
    py = -FocalLength * y_ / Divisor
End Sub
'*** End of cProjection


'*** ClassDef, describing a 3D-Quad-Structure which can recalculate and render itself on a 2D-Plane (a 2D-CairoContext)
Type c3DQuad 
    As Long FaceColor, BackFaceColor 
    Declare Constructor()
    Declare Constructor(P0 As t3D, P1 As t3D, P2 As t3D, P3 As t3D)
    Declare Function ReCalculateOn(P As cProjection) As Double
    Declare Sub DrawOn(CC As RC5_CairoContext, Alpha As Double=1, BackFaceStyle As eBackFaceStyle=Backface_greyed)
    Private:
    Const PC As Long = 4 'count of corner-points in this structure
    As t3D P3D(0 To PC - 1), N 'N being the Normal-Vector of this Quad
    As t2D P2D(0 To PC - 1) 
    As Double AvgDist, Depth, ND
End Type
Constructor c3DQuad()
End Constructor
Constructor c3DQuad(P0 As t3D, P1 As t3D, P2 As t3D, P3 As t3D)
    FaceColor = &H999999: BackFaceColor = &H333333 'ensure the Default-Colors
    P3D(0) = P0: P3D(1) = P1: P3D(2) = P2: P3D(3) = P3 'take over the Point-Info
    N = CrossV(SubV(P1, P2), SubV(P1, P0)) 'calculate the Normal-Vector
    Dim M As Double = MagV(N): If M Then N = NewV(N.x / M, N.y / M, N.z / M) 'normalize the Normal
End Constructor
Function c3DQuad.ReCalculateOn(P As cProjection) As Double 'ReCalculate on current Projection P
    P.Calc2DPt N.x, N.y, N.z, P2D(0).x, P2D(0).y, ND 'get the projected ND-Depth for our Normal-Vector
    AvgDist = 0
    For i As Long = 0 To UBound(P3D)
        P.Calc2DPt P3D(i).x, P3D(i).y, P3D(i).z, P2D(i).x, P2D(i).y, Depth
        AvgDist = AvgDist + Depth
    Next
    AvgDist = AvgDist / PC
    Return 10 - AvgDist 'return this value for the depth-sorting
End Function
Sub c3DQuad.DrawOn(CC As RC5_CairoContext, Alpha As Double=1, BackFaceStyle As eBackFaceStyle=Backface_greyed)
    Dim FillColor As Long, ShadeFac As Double
    If ND > 0 Then 'the normal Color- and Shade-Settings for the ForwardFace
        FillColor = FaceColor: ShadeFac = 0.2 + ND
    Else 'handle the backward-facing orientation of this Quad here
        If BackFaceStyle = BackFace_none Then Exit Sub 'do nothing at all in this case
        FillColor = FaceColor: ShadeFac = 0.2 - ND: Alpha = Alpha * 0.75 'generally give the backface a bit more transparency
        If BackFaceStyle = Backface_normal Then Alpha = Alpha * (1 - AvgDist) Else FillColor = BackFaceColor
    End If

    CC.PolygonPtr @P2D(0).x, PC, True 'simply draw the path of the (already calculated) 2D-PolyPoints (done by the Function above) 
    CC.SetSourceColor FillColor, Alpha, ShadeFac 'apply the just calculated Color-Settings
    CC.Fill 'and fill the Polygon
End Sub
'*** End of c3DQuad
 
'this routine can construct the z-Coord for different Area-Functions (the x/y Range of that area needs to be specified)
Sub FillAreaDataAs(FuncType As eFuncType, AreaInfo As tAreaInfo, xStart As Double, xEnd As Double, yStart As Double, yEnd As Double) 
  AreaInfo.FType = FuncType
  With AreaInfo
    Dim xStep As Double = (xEnd - xStart) / UBound(.Area3D, 1)
    Dim yStep As Double = (yEnd - yStart) / UBound(.Area3D, 2)
 
    For j As Integer = 0 To UBound(.Area3D, 2): For i As Integer = 0 To UBound(.Area3D, 1)
      With .Area3D(i, j)
        .x = xStart + i * xStep
        .y = yStart + j * yStep
        Select Case AreaInfo.FType '<- just enhance the Enum-Type and this Select about other interesting area-functions
          Case GaussianBell: .z = Exp(-(.x ^ 2 + .y ^ 2))
          Case Rosenbrock:   .z = (-.x ^ 2) + (.y - .x ^ 2) ^ 2
        End Select
      End With
    Next i, j
  End With
End Sub
 
'no matter what values are contained in a given AreaInfo-Struct, this routine will fit it into [-1,1]-ranges for x/y/z
Sub NormalizeAreaDataAndFillQuads(AreaInfo As tAreaInfo, Quads() As c3DQuad)
  With AreaInfo
    .xScale = 0: .yScale = 0: .zScale = 0
    
    'first pass to determine the Scale-Factors for each dimension
    Dim j As Long, i As Long, k As Long
    For j = 0 To UBound(.Area3D, 2): For i = 0 To UBound(.Area3D, 1)
      With .Area3D(i, j)
        If Abs(.x) > AreaInfo.xScale Then AreaInfo.xScale = Abs(.x)
        If Abs(.y) > AreaInfo.yScale Then AreaInfo.yScale = Abs(.y)
        If Abs(.z) > AreaInfo.zScale Then AreaInfo.zScale = Abs(.z)
      End With
    Next i, j
    
    'to avoid potential Div by Zero-Errors within the loop that follows...
    If .xScale = 0 Then .xScale = 1E-50
    If .yScale = 0 Then .yScale = 1E-50
    If .zScale = 0 Then .zScale = 1E-50
   
    'second pass to apply the found scale-factors, normalizing any Dimension into the [-1 to +1] range
    For j = 0 To UBound(.Area3D, 2): For i = 0 To UBound(.Area3D, 1)
      With .Area3D(i, j)
        .x = .x / AreaInfo.xScale
        .y = .y / AreaInfo.yScale
        .z = .z / AreaInfo.zScale
      End With
    Next i, j
    
    'third pass to determine the Min/Max-Values for each dimension on the already scaled values
    .xMin = 1E+50: .xMax = -1E+50
    .yMin = 1E+50: .yMax = -1E+50
    .zMin = 1E+50: .zMax = -1E+50
  
    For j = 0 To UBound(.Area3D, 2): For i = 0 To UBound(.Area3D, 1)
      With .Area3D(i, j)
        If .x < AreaInfo.xMin Then AreaInfo.xMin = .x
        If .x > AreaInfo.xMax Then AreaInfo.xMax = .x
        If .y < AreaInfo.yMin Then AreaInfo.yMin = .y
        If .y > AreaInfo.yMax Then AreaInfo.yMax = .y
        If .z < AreaInfo.zMin Then AreaInfo.zMin = .z
        If .z > AreaInfo.zMax Then AreaInfo.zMax = .z
      End With
    Next i, j
    
    'finally add the Quads
    ReDim Quads(0 To UBound(.Area3D, 1) * UBound(.Area3D, 2) - 1)
    For j = 1 To UBound(.Area3D, 2): For i = 1 To UBound(.Area3D, 1) 'we start at Indexes 1 (leaving out the Zero-Indexes)
      Quads(k) = c3DQuad(.Area3D(i-1, j-1), .Area3D(i-1, j), .Area3D(i, j), .Area3D(i, j-1))
      If .zMax - .zMin > 0 Then Quads(k).FaceColor = Spectrum.LUT((.Area3D(i-1, j-1).z - .zMin) * UBound(Spectrum.LUT) / (.zMax - .zMin))
      k = k + 1
    Next i, j
  End With
End Sub

'a unit-cube can be constructed without any Math (so we fill in its 6 sides into a c3DQuad-Array directly here)
Sub ConstructUnitCubeAndFillQuads(AreaInfo As tAreaInfo, Quads() As c3DQuad) 
    AreaInfo.FType = UnitCube
    AreaInfo.xMin = -1: AreaInfo.xMax = 1: AreaInfo.xScale = 1
    AreaInfo.yMin = -1: AreaInfo.yMax = 1: AreaInfo.yScale = 1
    AreaInfo.zMin = -1: AreaInfo.zMax = 1: AreaInfo.zScale = 1

    ReDim Quads(0 To 5)
    Quads(0) = c3DQuad(NewV(-1,-1,-1), NewV(-1,-1,+1), NewV( 1,-1,+1), NewV( 1,-1,-1))
    Quads(1) = c3DQuad(NewV(-1,-1,+1), NewV(-1,-1,-1), NewV(-1, 1,-1), NewV(-1,+1,+1))
    Quads(2) = c3DQuad(NewV(-1,+1,-1), NewV(-1,-1,-1), NewV(+1,-1,-1), NewV(+1,+1,-1))
    Quads(3) = c3DQuad(NewV(-1,-1,+1), NewV(-1,+1,+1), NewV(+1,+1,+1), NewV(+1,-1,+1))
    Quads(4) = c3DQuad(NewV(-1,+1,+1), NewV(-1,+1,-1), NewV(+1,+1,-1), NewV(+1,+1,+1))
    Quads(5) = c3DQuad(NewV(+1,-1,-1), NewV( 1,-1, 1), NewV(+1,+1,+1), NewV(+1,+1,-1))
End Sub